(*
	MIDIPLAY.C

	8-May-89, Dale Glowinski, Ad Lib Inc.

	Driver to play a Standard MIDI File (1.0).

	The driver uses the PC's timer-0 for timing. (see TIMER.ASM )

	This file was compiled with Microsoft C5.1 as follows:
	   cl -AL -Zi -J -Ot -Zp -Gs -c -Foobject\\
	The companion file, TIMER.ASM, was assembled with MASM 5.1:
	   masm -Zi timer.asm, object\\timer.obj;

	Interface instructions:
	   - Call Midi_Init() to install the interrupt vector.
	   - Call Midi_Play (*data) to play the file.
		 *data is a pointer to the already read file
	   - Call Midi_End() to remove the interrupt vector.

	   - You do not need to call Midi_Init and Midi_End every time you
		 play something.  You can call them once at the beginning and
		 end of your program.
	   - The timer-0 is adjusted according to the tempo to adjust the clock
		 interrupt frequency.  The new interrupt routine takes care of
		 chaining to the old interrupt routine at the appropriate moment.
		 Timer-0 is set to its usual state when finished playing.
	   - See the file midimain.c for an example.
*)

{#include  "cflags.h"

#include  <stdio.h>
#include  <stdlib.h>
#include  <fcntl.h>
#include  <string.h>
}

UNIT MIDIPLAY;

INTERFACE

USES
    ADLIBU;

{$i cflags.inc}

(* Some MIDI codes. *)

CONST
     END_OF_TRACK =  $2f;
     TEMPO        =  $51;
     NR_CHANS     =  $16;
     SYSEX_F0     =  $f0;
     SYSEX_F7     =  $f7;
     META         =  $ff;
     SEQ_SPECIFIC =  $7f;


VAR
   tracks,                 { Number of tracks }
   tickQnote : INTEGER;    { ticks per quarter note }
   musPtrPtr : ^BPTR;      { ptr to trackptrs }
   status   : ^BYTE;       { ptr to running status of current track }

CONST
     clock_rate : WORD = 0;     { Original clock interrupt rate }
     tickCount  : LONGINT = 0;  { Current absolute time }
     ctrk       : INTEGER = 0;  { Current track }

CONST
     TTRACKS = 16;

TYPE
    BPTR = ^BYTE;
    TRACK = ARRAY [0..TTRACKS] OF ^BYTE;

VAR
   trkPtrs : ARRAY [0..TTRACKS] OF ^BYTE;   { ptrs to each data track }
   trkStats: ARRAY [0..TTRACKS] OF BYTE;    { running status for each track }
   abs_time: ARRAY [0..TTRACKS] OF LONGINT; { time of next event for each track }


(* Keeps track of last change of volume to avoid unnecessary calls to
   change of volume routine. *)

VAR
   current_vol : ARRAY [0..TTRACKS] OF INTEGER;

CONST
     volume_flag : INTEGER = 1;

(* Flags *)

CONST
     musRunning : BOOLEAN = FALSE;
     end_of_data : BOOLEAN = FALSE;

CONST
     clock_in : BOOLEAN = FALSE;

(* Prototypes *)

{ SetUp_Data (UCHAR *); }
{ static Start_Melo (); }
{ static unsigned  Do_Event ();}

{
#ifdef INT_METHOD
   unsigned TimeOut ();
#endif
}

PROCEDURE Midi_Init;
FUNCTION Midi_Play (dataPtr:BPTR) : BOOLEAN;
PROCEDURE Midi_End;
PROCEDURE Stop_Melo;
PROCEDURE Set_Original_Clock;
PROCEDURE Set_Tempo (tickQnote: WORD; usec: LONGINT);
FUNCTION  TimeOut : WORD;
PROCEDURE Midi_Kill;

{$ifndef INT_METHOD}
FUNCTION Test_Event : INTEGER;
{$endif}

IMPLEMENTATION

FUNCTION Get_Word (ptr : BPTR) : WORD; FORWARD;
FUNCTION Get_Long (ptr : BPTR) : LONGINT; FORWARD;
PROCEDURE SetUp_Tracks (trcks : INTEGER; chunk : BPTR); FORWARD;
FUNCTION Get_Length : LONGINT; FORWARD;
PROCEDURE start_Melo; FORWARD;
FUNCTION Do_Event : WORD; FORWARD;

PROCEDURE Clk_Uninstall; EXTERNAL;               {$L TIMER}
PROCEDURE StartTimeout (x : INTEGER); EXTERNAL;
PROCEDURE SetClkRate (x : INTEGER); EXTERNAL;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/   Set up all of the data structures used in playing a MIDI file        _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

PROCEDURE SetUp_Data (dataPtr : BPTR);

VAR
   length : LONGINT;
   i, j   : INTEGER;
   dptr4, dptr10, dptr12 : BPTR;

BEGIN
     FOR I:=0 TO MAX_VOICES-1 DO
     BEGIN
          current_vol [i] := 0;
          SetVoiceVolume (i, 0);
          NoteOff (i);
     END;

     tickCount := 0;             { current absolute time }
     ctrk := 0;                  { current track }

     { Read file header }

     dptr4  := dataPtr;     INC (dptr4 ,4 );
     dptr10 := dataPtr;     INC (dptr10,10);
     dptr12 := dataPtr;     INC (dptr12,12);

     length    := Get_Long (dptr4);      { header length }
     tracks    := Get_Word (dptr10);     { number of tracks }
     tickQnote := Get_Word (dptr12);     { ticks per quarter note }

     { Set musPtr to point to start of first chunk }

     INC (dataPtr, length + 8);

     { Set up the array trkPtrs }

     SetUp_Tracks (tracks, dataPtr);

     { Initialize arrays }

     FOR i:=0 TO tracks-1 DO
     BEGIN
	  musPtrPtr := ADDR(trkPtrs [i]);    { set global data ptr }
	  abs_time [i] := Get_Length;        { set start time for track }
	  trkStats [i] := musPtrPtr^^;       { current running status for track }
     END;
END;



PROCEDURE Clk_install; EXTERNAL;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                    _/}
{_/	Install the clock interrupt routine.                           _/}
{_/                                                                    _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

PROCEDURE Midi_Init;

BEGIN
     IF (clock_in) THEN EXIT;
     Clk_install;
     clock_in := TRUE;
END;


{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/Main routine for playing a MIDI file.  It receives an array of function _/}
{_/pointers which are used to call the routines which process the MIDI     _/}
{_/events.                                                                 _/}
{_/Returns 0 if interrupt routine not installed, else returns 1.           _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

FUNCTION Midi_Play (dataPtr:BPTR) : BOOLEAN;

BEGIN
     IF (NOT clock_in) THEN
     BEGIN
          Midi_Play := FALSE;
          EXIT;
     END;
     SetUp_Data (dataPtr);
     Start_Melo;
     Midi_Play := TRUE;
END;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/      Uninstall the clock driver...                                     _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

PROCEDURE Midi_End;

BEGIN
     IF (clock_in) THEN Clk_uninstall;
     clock_in := FALSE;
END;


{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/      Get word value from data.  Value is stored MSB first              _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}


FUNCTION Get_Word (ptr : BPTR) : WORD;

VAR
   lsb,msb : BYTE;

BEGIN
     msb := ptr^;   INC (ptr);
     lsb := ptr^;
     Get_word := lsb + (msb SHL 8);
END;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/      Get a long value from data.  Value is stored big endian           _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}


FUNCTION Get_Long (ptr : BPTR) : LONGINT;

VAR
   msw : WORD;
   lsw : WORD;

BEGIN
     msw := Get_Word (ptr);      INC (ptr,2);
     lsw := Get_Word (ptr);
     Get_Long := LONGINT(lsw) + (LONGINT(msw) SHL 16);
END;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/      Set up trkPtrs, which is an array of pointers, to point to the    _/}
{_/      track chunks.  Does not modify musPtr                             _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

PROCEDURE SetUp_Tracks (trcks : INTEGER; chunk : BPTR);

VAR
   n : INTEGER;
   length : LONGINT;
   tPtr : ^TRACK;


BEGIN
     tPtr := ADDR(trkPtrs);

     FOR n:=0 TO trcks-1 DO
     BEGIN
          INC (chunk,4);
          length   := Get_Long (chunk);
          INC (chunk,4);
          tPtr^ [n] := POINTER(chunk);
     END;
END;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/      Reads a variable length value from the MIDI file data and advances_/}
{_/      the data pointer                                                  _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}


FUNCTION Get_Length : LONGINT;

VAR
   value : LONGINT;
   c : BYTE;
   data : BPTR;

BEGIN
     data := BPTR(musPtrPtr^);
     value := LONGINT(data^);
     INC (data);

     IF (value AND $80 = $80) THEN
     BEGIN
           value := value AND $7f;
           REPEAT
                 c := data^;
                 INC (data);
                 value := (value SHL 7) + (c AND $7f);
           UNTIL (c < 127);
     END;

     musPtrPtr^ := POINTER(data);
     Get_Length := value;
END;


{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/   Start playing a melody.  Set some global pointer and the temp. Start _/}
{_/   the clock driver with the first delay (>=1)                          _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

PROCEDURE start_Melo;

BEGIN
     musPtrPtr   := ADDR(trkPtrs);
     status      := ADDR(trkStats);
     end_of_data := FALSE;
     Set_Tempo (0, 500000);
     StartTimeOut (Do_Event);
     musRunning  := TRUE;
END;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/   Stop playing the melody.  Reset the clock frequency to normal (18.2) _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

PROCEDURE Stop_Melo;

BEGIN
     musRunning := FALSE;
     Set_Original_Clock;
END;

PROCEDURE Midi_Kill;
VAR
   n : INTEGER;
BEGIN
     IF (MusRunning) THEN
     BEGIN
          FOR n := 0 TO MAX_VOICES-1 DO
          BEGIN
               SetVoiceVolume(n,0);
               NoteOff(n);
          END;
          Stop_Melo;
     END;
END;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/  Set clock rate to its original interrupt rate. Note that the clock rate/}
{_/  has been saved at 10 times its real value in order to preserve some   _/}
{_/  accuracy.                                                             _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

PROCEDURE Set_Original_Clock;

BEGIN
     SetClkRate (16*256+10);
END;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/  Change the tempo                                                      _/}
{_/                                                                        _/}
{_/  Reload the timer-0 with the proper divisor for generating             _/}
{_/  the appropriate frequency.                                            _/}
{_/                                                                        _/}
{_/  If tempo is zero, reprogram the counter for 18.2 Hz                   _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

PROCEDURE Set_Tempo (tickQnote: WORD; { ticks per quarter note }
                    usec: LONGINT);   { micro-seconds per quarter note }
VAR
   count : LONGINT;

BEGIN
     IF (tickQnote=0) THEN count := 0
     ELSE
     BEGIN
          usec := usec DIV 1000;
          count := 1194 * usec DIV tickQnote;
     END;
     SetClkRate (count);
END;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/  Finds the next event to be processed by checking 'abs_times', which   _/}
{_/  contains the absolute time (in ticks) of when the next event occurs   _/}
{_/  for each track. Sets the global data ptr musPtr to point to the next  _/}
{_/  event and sets the running status for that track as well.             _/}
{_/  Returns number of ticks until next event. */                          _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

FUNCTION Get_Next_Delay : WORD;

VAR
   delta : LONGINT;
   n, min : INTEGER;

BEGIN
     IF (status^ <> END_OF_TRACK) THEN
     BEGIN
	  delta := Get_Length;                 { get delta time }
	  INC (abs_time [ctrk], delta);        { set new time for track }
     END
     ELSE
         abs_time [ctrk] := $7fffffff;         { impossibly large value }

     { Find earliest time in abs_time array.  This determines which track
	  contains the next event. }

     min := 0;
     FOR n:=1 TO tracks-1 DO
         IF ((abs_time [n] < abs_time [min]) AND (trkStats [n] <> END_OF_TRACK)) THEN
	    min := n;

     IF (trkStats [min] = END_OF_TRACK) THEN
     BEGIN
	  { end of data condition for all tracks }
	  end_of_data := TRUE;
	  Stop_Melo;
          Get_Next_Delay := 0;
          EXIT;
     END;

     delta     := abs_time [min] - tickCount; { calculate time until next event }
     tickCount := abs_time [min];             { set current time }
     musPtrPtr := ADDR(trkPtrs [min]);        { reset data ptr }
     status    := ADDR(trkStats [min]);       { set running status }
     ctrk      := min;

     Get_Next_Delay := delta;
END;

PROCEDURE myNoteOn ( voice, note, volume: INTEGER);

BEGIN
     IF (volume = 0) THEN
     BEGIN
	  { A note-on with a volume of 0 is equivalent to a note-off. }
	  NoteOff (voice);
	  current_vol [voice] := volume;
     END
     ELSE
     BEGIN
	  { Regular note-on }
	  IF (current_vol [voice] <> volume) THEN
          BEGIN
	       SetVoiceVolume (voice, volume);
	       current_vol [voice] := volume;
          END;
	  NoteOn (voice, note);
     END;
END;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/  Process a regular MIDI event.  Which routine to call is determined by _/}
{_/  using the 3 LSB's of the high nibble                                  _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

PROCEDURE Midi_Event (event:WORD);


   { Table of # of data bytes which follow a regular midi status byte }

CONST
     data_bytes : ARRAY [0..6] OF BYTE = ( 2, 2, 2, 2, 1, 1, 2 );

VAR
   stat, voice : INTEGER;
   mus1 , musPtr: BPTR;

BEGIN
     stat  := (event SHR 4) AND 7;
     voice := event AND $0f;
     musptr := BPTR(musPtrPtr^);
     mus1 := musPtr;  INC (mus1);

     IF (voice < MAX_VOICES) THEN
        CASE (stat) OF
	     0:NoteOFF (voice);
	     1:myNoteOn (voice, musPtr^, mus1^);
	     2:BEGIN
                    IF ((current_vol [voice] <> mus1^) AND (volume_flag=1)) THEN
                       SetVoiceVolume (voice, mus1^);
		       current_vol [voice] := mus1^;
                    END;
	     5:BEGIN
		    IF (current_vol [voice] <> musPtr^) THEN
		    BEGIN
                         IF (volume_flag=1) THEN
                            SetVoiceVolume (voice, musPtr^);
		         current_vol [voice] := musPtr^;
                    END;
               END;
	     6:SetVoicePitch (voice, ((mus1^) SHL 7) OR musPtr^);
        END;

        INC (musPtrPtr^, data_bytes [stat]);
END;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/  Proces an Ad Lib specific meta-event.                                 _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}


PROCEDURE AdLib_Specific (code:INTEGER; data:BPTR);

VAR
   n : INTEGER;
   parms : ARRAY [0..27] OF INTEGER;
   aptr : BPTR;

BEGIN
     IF (code = 1) THEN
     BEGIN

	  { Instrument change code.  First byte of data contains voice number.
		 Following bytes contain instrument parameters.  }

          FOR n:=0 TO 27 DO
          BEGIN
               aptr := data;    INC (aptr,n+1);
               parms [n] := aptr^ AND $ff;
          END;
          SetVoiceTimbre (INTEGER(data^), ADDR(parms));
     END
     ELSE
         IF (code = 2) THEN
	      { Melo/perc mode code.  0 is melodic, !0 is percussive. }

              SetMode (data^=1)

         ELSE
         IF (code = 3) THEN
   	    SetPitchRange (INTEGER(data^));
END;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/ Proces meta-event.  All events other than end-of-track and tempo events_/}
{_/ are ignored                                                            _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

PROCEDURE Meta_Event;

TYPE
    array5 = ARRAY [0..5] OF BYTE;

VAR
   l : LONGINT;
   data : ^array5;

BEGIN
     { musPtr points to the event type byte which follows the 0xff. }
     if (musPtrPtr^^ = END_OF_TRACK) THEN
     BEGIN
	  status^ := END_OF_TRACK;
          DEC (musPtrPtr^);                      { leave ptr on EOT event }
     END
     ELSE
         IF (musPtrPtr^^ = TEMPO) THEN
         BEGIN
              INC (musPtrPtr^,2);
              l := musPtrPtr^^;              INC (musPtrPtr^);
              l := (l SHL 8) + musPtrPtr^^;  INC (musPtrPtr^);
              l := (l SHL 8) + musPtrPtr^^;  INC (musPtrPtr^);
	      Set_Tempo (tickQnote, l);
         END
         ELSE
             IF (musPtrPtr^^ = SEQ_SPECIFIC) THEN
             BEGIN
                  INC (musPtrPtr^);          { event type byte }
                  l := Get_Length;
                  data := POINTER(musPtrPtr^);

	          { Ad Lib midi ID is 00 00 3f. }

	          IF ((data^ [0] = 0) AND (data^ [1] = 0) AND (data^ [2] = $3f)) THEN

		  { The first two bytes after the ID contain the Ad Lib event code.
			The following bytes contain the data pertaining to the event. }

		     AdLib_Specific ((data^ [3] SHL 8) OR (data^ [4]), ADDR(data^ [5]));

	          INC (musPtrPtr^,l);
             END
             ELSE
             BEGIN
                  INC (musPtrPtr^);
                  INC (musPtrPtr^,Get_Length);
             END;
END;

{-------------------------------------------------------------------------}

PROCEDURE Sysex_Event (event:BYTE);

VAR
   len : LONGINT;

BEGIN
     len := Get_Length;

     { skip over system exclusive event }
     INC (musPtrPtr^,len);
END;


{$ifdef INT_METHOD}

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/	Interrupt routine. Called by low-level clock driver when           _/}
{_/	the delay count has expired.                                       _/}
{_/                                                                        _/}
{_/	'musPtr' always points to the first byte AFTER the timing byte.    _/}
{_/                                                                        _/}
{_/	When this routine is called, the active SS  (stack segment) is not _/}
{_/	the original of the application, so take care.                     _/}
{_/	This routine, and all others called by, must be compiled           _/}
{_/	without stack-overflow checking, since the SS has changed!!!       _/}
{_/                                                                        _/}
{_/	Return to caller the number of clock ticks to wait for before      _/}
{_/	the next call.                                                     _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

FUNCTION TimeOut : WORD;

BEGIN
     IF (NOT musRunning) THEN
        { Music has not started or has been stopped, so wait the minimum delay ... }
        TimeOut := 1
     ELSE
         TimeOut := Do_Event;
END;

FUNCTION Do_Event : WORD;

VAR
   delay : WORD;

BEGIN
     REPEAT
           { If high bit set, set status, else this is running status. }
 	   IF ((musPtrPtr^^ AND $80) = $80) THEN
           BEGIN
	        status^ := musPtrPtr^^;
                INC (musPtrPtr^);
           END;

	   { Process event. }

           IF ((status^ = SYSEX_F7) OR (status^ = SYSEX_F0)) THEN
	      Sysex_Event (status^)
           ELSE
	       IF (status^ = META) THEN
                  Meta_Event
	       ELSE
                   Midi_Event (status^);

		{ Read next delta time. }

	   delay := Get_Next_Delay;
     UNTIL ((delay <> 0) OR end_of_data);

     IF (delay = 0) THEN Do_Event := 1
     ELSE Do_Event := delay;
END;

{$else}

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/  Alternate method: instead of all the work being done by the interrupt _/}
{_/  routine, the interrupt routine sets a flag and exits.  The main       _/}
{_/  program loops on this flag and processes an event whenever the flag   _/}
{_/  is set.  The main program then resets the flag and tells the interrupt_/}
{_/  routine how long to wait until setting the flag again.                _/}
{_/  The avantage of this is that it can be used for debugging.  Codeview  _/}
{_/  cannot be used for debugging interrupt routines, for example.  As well_/}
{_/  using "printf" inside an interrupt routine will cause the system to   _/}
{_/  hang.                                                                 _/}
{_/  The avantage of the first method is that you can go off and do other  _/}
{_/  things while the music is playing.  With this method, you must watch  _/}
{_/  the flag constantly.  Use this method to debug and then switch to the _/}
{_/  interrupt method for your final version.                              _/}
{_/  To use this method, go into the file "cflags.h", put the define for   _/}
{_/  INT_METHOD in comments and recompile midimain.c and this file.        _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}


CONST
     timer_signal : BOOLEAN = FALSE;


{ Interrupt routine. Called by low-level clock driver when
   the delay count has expired.

   When this routine is called, the active SS  (stack segment) is not
   the original of the application, so take care.
   This routine, and all others called by, must be compiled
   without stack-overflow checking, since the SS has changed!!!
}

FUNCTION  TimeOut : WORD;

BEGIN
     timer_signal := TRUE;
     TimeOut := $7fff;
END;

{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}
{_/                                                                        _/}
{_/   Processes the next event and returns the delay until the next event  _/}
{_/   occurs.                                                              _/}
{_/   'musPtr' always points to the first byte AFTER the timing byte.      _/}
{_/                                                                        _/}
{_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/}

FUNCTION Do_Event : WORD;

VAR
   delay : WORD;

BEGIN
     timer_signal := FALSE;

     IF (NOT musRunning) THEN
     BEGIN
          Do_Event := 1;
          EXIT;
     	  { Music has not started or has been stopped, so wait the minimum delay ... }
     END;

     REPEAT
	   { If high bit set, set status, else this is running status. }
	   IF (musPtrPtr^^ > 127) THEN
           BEGIN
	        status^ := musPtrPtr^^;
                INC (musPtrPtr^);
           END;

	   { Process event. }

	   IF ((status^ = SYSEX_F7) OR (status^ = SYSEX_F0))
           THEN
	       Sysex_Event (status^)
           ELSE
	       IF (status^ = META) THEN Meta_Event
               ELSE
	           Midi_Event (status^);

	       { Read next delta time. }

	       delay := Get_Next_Delay;
     UNTIL ((delay <> 0) OR end_of_data);

     IF (delay = 0) THEN
        Do_Event := 1
     ELSE
         Do_Event := delay;
END;

FUNCTION Test_Event : INTEGER;

BEGIN
     IF (NOT timer_signal) THEN EXIT;
     StartTimeOut (Do_Event);
END;

{$endif}

END.